#ifndef __NEURALNETWORK_H
#define __NEURALNETWORK_H

#include <vector>
#include <valarray>

#include "../Controller.h"
#include "../GenotypeDecoder.h"
#include "../../utils/Graph.h"
#include "Neuron.h"

/*!	\file GraphNeuralNetwork.h
	\brief Interface of neural networks based on graphs
*/

namespace Teem
{
	//! A generic neural network with arbitrary topology and neurons/synapses types.
	//! \ingroup controllers
	class GraphNeuralNetwork :  public Controller
	{
	public:
		//! Constructor, create this neural network using graph as a description
		GraphNeuralNetwork(const Graph &graph, const std::string &root = "teem.nn");
		//! Destructor, delete neurons and synapses
		virtual ~GraphNeuralNetwork();
		
		// Generic controller methods
		virtual void setInput(unsigned index, double val);
		virtual double getOutput(unsigned index);
		virtual void step();
		
		// Accessors.
		//! Get a neuron by index.
		Neuron* getNeuron(size_t index) { return neurons[index]; }
		//! Get a neuron from a tagged list by index.
		Neuron* getNeuron(const std::string &tag, size_t index) { return neuronLists[tag][index]; }
		//! Get an array of tagged neurons.
		std::valarray<Neuron*>& getTaggedNeurons(const std::string &tag) { return neuronLists[tag]; }
		//! Total number of neuron.
		size_t neuronNum() const { return neurons.size(); }
		//! Number of neuron in a tagged list.
		size_t neuronNum(const std::string &tag) const { return neuronLists.find(tag)->second.size(); }
		//! Get a synapse by index.
		Synapse* getSynapse(size_t index) { return synapses[index]; }
		//! Get a synapse from a tagged list by index.
		Synapse* getSynapse(const std::string &tag, size_t index) { return synapseLists[tag][index]; }
		//! Get an array of tagged synapses.
		std::valarray<Synapse*>& getTaggedSynapses(const std::string &tag) { return synapseLists[tag]; }
		//! Total number of synapse.
		size_t synapseNum() const { return synapses.size(); }
		//! Number of synapse in a tagged list.
		size_t synapseNum(const std::string &tag) const { return synapseLists.find(tag)->second.size(); }
		
	protected:
		//! All neurons
		std::valarray<Neuron*> neurons;
		//! All synapses
		std::valarray<Synapse*> synapses;
		
		//! Neuron tagged lists.
		std::map<std::string, std::valarray<Neuron*> > neuronLists;
		//! Synapse tagged lists.
		std::map<std::string, std::valarray<Synapse*> > synapseLists;
		
		//! Cached input neurons (for performances). This is a copy of \c neuronLists["input"] with
		//! object already dynamic_casted.
		std::valarray<InputNeuron*> inputNeurons;
		//! Cached output neurons (for performances). This is just a pointer to \c neuronLists["output"].
		std::valarray<Neuron*>* outputNeurons;
	};
	
	
	/**
	This class implements decoding of a feedforward graph neural network.
	Currently, this class handles only sigmoid neural network.
	TODO: generalize this class for arbitrary parameters and arbitrary
	neuron types.
	\ingroup genotypeDecoders
	*/
	class GraphFeedForwardGenotypeDecoder : public GraphGenotypeDecoder
	{
	protected:
		Ishtar::Variable<unsigned> hiddenCount; //!< Number of neurons in the hidden layer
		Ishtar::Variable<double> weightRange; //!< Range of synapse's weights, which will be within [-weightRange .. weightRange]

	public:
		//! Constructor, by default create no hidden neuron and set weight range to 2
		GraphFeedForwardGenotypeDecoder(const std::string &root) :
			GraphGenotypeDecoder(root),
			hiddenCount(root + ".hiddenCount", 0),
			weightRange(root + ".weightRange", 2.0) { }
		//! Virtual destructor, do nothing
		virtual ~GraphFeedForwardGenotypeDecoder() { }

		// GenotypeDecoder overrides
		virtual Genome* createGenome(void);
		virtual Controller* decode(Genome *genome);
	};

	/**
	This class implementes decoding of a fully connected graph neural network.
	Currently, this class handles only sigmoid neural network.
	TODO: generalize this class for arbitrary parameters and arbitrary
	neuron types.
	\ingroup genotypeDecoders
	*/
	class GraphFullyConnectedGenotypeDecoder : public GraphGenotypeDecoder
	{
	protected:
		Ishtar::Variable<unsigned> hiddenCount; //!< Number of neurons in the hidden layer
		Ishtar::Variable<double> weightRange; //!< Range of synapse's weights, which will be within [-weightRange .. weightRange]

	public:
		//! Constructor, by default create no hidden neuron and set weight range to 2
		GraphFullyConnectedGenotypeDecoder(const std::string &root) :
			GraphGenotypeDecoder(root),
			hiddenCount(root + ".hiddenCount", 0),
			weightRange(root + ".weightRange", 2.0) { }
		//! Virtual destructor, do nothing
		virtual ~GraphFullyConnectedGenotypeDecoder() { }
		
		// GenotypeDecoder overrides
		virtual Genome* createGenome(void);
		virtual Controller* decode(Genome *genome);
	};

}

#endif
