#include "Neuron.h"
#include "../../main/VerboseLevel.h"
#include <assert.h>
#include <math.h>
#include <iomanip>

/*!	\file Neuron.cpp
	\brief Implementations of neurons and synapses
*/

namespace Teem
{
	Neuron::Neuron()
	{
		lastSpike = false;
		lastCurrent = 0.0;
	}
	
	Neuron::~Neuron()
	{
	}
	
	void Neuron::setParams(const std::map<std::string, double> params)
	{
		for(std::map<std::string, double>::const_iterator i = params.begin(); i != params.end(); i++)
			setParams(i->first, i->second);
	}
	
	void Neuron::setParams(const std::string &param, double value)
	{
		COUT(2, "Neuron::setParams : unhandled parameter \"" << param << "\" with value " << value << "." << std::endl);
	}
	
	void InputNeuron::setParams(const std::string &param, double value)
	{
		if(param == "value")
			setInputValue(value);
		else	
			Neuron::setParams(param, value);
	}
	
	void FiringRateOutputNeuron::setParams(const std::string &param, double value)
	{
		if(param == "timeWindow")
			queueSize = unsigned(floor(value+0.5));	// = round(value), but round is *not* supported by real sbots
		else
			Neuron::setParams(param, value);
	}
	
	void FiringRateOutputNeuron::update(double dt)
	{
		// count incoming spikes
		unsigned spikeCount = 0;
		for (SynapseSet::iterator p = incoming.begin(); p != incoming.end(); ++p)
			spikeCount += (*p)->getSpike() ? 1 : 0;
		
		// save spike count.
		totalSpike += spikeCount;
		spikes.push(spikeCount);
		
		while(spikes.size() > queueSize)
		{
			totalSpike -= spikes.front();
			spikes.pop();
		}
	}
	
	void FiringRateOutputNeuron::commit()
	{
		lastCurrent = ((double) totalSpike) / ((double) spikes.size() * incoming.size());
	}
	
	Synapse::Synapse()
	{
		preNeuron = NULL;
		postNeuron = NULL;
	}
	
	void Synapse::init(Neuron* pre, Neuron* post, const double weight)
	{
		assert(pre);
		assert(post);
		assert(finite(weight));
		
		this->weight = weight;
		preNeuron = pre;
		postNeuron = post;
	
		preNeuron->outgoing.insert(this);
		postNeuron->incoming.insert(this);
	}
	
	Synapse::~Synapse()
	{
		preNeuron->outgoing.erase(this);
		postNeuron->incoming.erase(this);
	}
	
	void Synapse::setParams(const std::map<std::string, double> params)
	{
		for(std::map<std::string, double>::const_iterator i = params.begin(); i != params.end(); i++)
			setParams(i->first, i->second);
	}
	
	void Synapse::setParams(const std::string &param, double value)
	{
		if(param == "weight")
			weight = value;
		else
			COUT(2, "Synapse::setParams : unhandled parameter \"" << param << "\" with value " << value << "." << std::endl);
	}
}
