#ifndef __ENKI_VIEWER_H
#define __ENKI_VIEWER_H

#ifdef HAVE_VIEWER

#include <string>

#include <ishtar/settings.h>
#include <enki/PhysicalEngine.h>

#include "videorec.h"

#include "../libgag/GAG.h"

namespace Enki
{
	class Alice;
}

//! Viewer that uses libgag to draw things
class Viewer
{
protected:
	GAG::GraphicContext *gc; //!< libgag's graphic context
	GAG::Font *font; //!< font for drawing text
	unsigned eventStep; //!< step counter to skip most steps when display is disable
	unsigned delayStartTick; //!< timestamp of the beginning of step
	bool displayAll; //!< true if each step has to be displayed
	#ifdef HAVE_AVCODEC
	VideoRecorder *vrec; //!< pointer to video recorder. If NULL no video recorder
	#endif
	//! Width and height of the window
	int width, height;
	
	Ishtar::Variable<unsigned> realtime; //!< is realtime mode enabled
	Ishtar::Variable<std::string, true> fontFileName; //!< filename of the font
	
	//! Display the viewer
	virtual void displayViewer(int w, int h) = 0;

public:
	//! Constructor, by default do not record videos
	Viewer(bool recVideo = false);
	//! Destructor
	virtual ~Viewer();
	
	//! Create a window of size width x height
	void init(int width, int height);
	//! Do a step: display, sleep for dt if realtime and handle events
	void step(double dt);
	
	//! Process incoming events
	void handleEvent(bool eachStep = false);
	//! Beginning recording of the video named fileName
	void startRecordingVideo(const std::string &fileName);
	//! Stop the recording of the video, close the file
	void stopRecordingVideo(void);
	//! Display and record frame if video is active. If force is true, disregard displayAll and display anyway
	void display(bool force = false);
	//! Set the caption of the window
	void setCaption(const std::string &caption);
	//! Record timestamp for delay computation
	void startDelayFrame(void);
	//! Make sure that at least dt second have passed since last startDelayFrame call. If necessary, sleep
	void endDelayFrame(double dt);
};

//! A viewer that draw Enki worlds
//! Warning: not very clean hack
class EnkiViewer : public Viewer
{
protected:
	const Enki::World *world; //!< pointer to world in Enki
	
	Ishtar::Variable<double> scale; //!< scaling factor of the world (pixel per cm in Enki)
	int sideBarWidth; //!< width of the sidebar (used by subclasses to draw additional informations)
	
public:
	//! Constructor, use world w
	EnkiViewer(const Enki::World *w, bool recVideo = false, int sideBarWidth = 0);
	//! Destructor, do not touch world
	virtual ~EnkiViewer() { }

protected:
	//! Draw everything
	virtual void displayViewer(int w, int h);
	//! Draw the world content
	virtual void displayWorld(int x, int y, int w, int h);
	//! Draw the side bar content, by default do nothing
	virtual void displaySideBar(int x, int y, int w, int h);
	// not Antoine's hack!
	//! Draw Alice's IR sensors
	void drawIRSensors (int bottom, const Enki::Alice *owner, int which);
	//! Draw Alice's camera
	void drawCameraRays(int bottom, const Enki::Alice *owner);

};

//! Initial hack at a class that displays vision
//! Warning: not very clean hack
class VisionEnkiViewer : public EnkiViewer
{
public:
	//! Create a vision viewer for world w capable of displaying robot camera
	VisionEnkiViewer(const Enki::World *w, bool recVideo = false, int sideBarWidth = 268) : EnkiViewer(w, recVideo, sideBarWidth) { }
	//! Destructor
	virtual ~VisionEnkiViewer() { };
	
protected:
	//! Draw camera in the side bar
	virtual void displaySideBar(int x, int y, int w, int h);
	//! Draw a ray corresponding to a sensor
	void drawIRSensor(double dx, double dy, double angle, int bottom, const Enki::PhysicalObject &owner);
};

#endif // HAVE_VIEWER

#endif 
