#ifdef HAVE_AVCODEC

#include "videorec.h"
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#include <iostream>

struct Vid
{
	AVFormatContext oc;
	AVStream *vs;
	AVFrame *toEncodeFrame;
	int outbufSize;
	unsigned char *outbuf;
	bool ok;
} __vid;


// Some utilities

AVFrame *allocFrame(int pix_fmt, int width, int height)
{
	AVFrame *picture = avcodec_alloc_frame();
	if (!picture)
		return NULL;
		
	int size = avpicture_get_size(pix_fmt, width, height);
    if (size < 0)
		return NULL;
	
	void *ptr = av_malloc(size);
	if (!ptr)
		return NULL;
		
	avpicture_fill((AVPicture *)picture, (unsigned char *)ptr, pix_fmt, width, height);
	if (!picture)
		return NULL;
		
	return picture;
}

// The class impl

VideoRecorder::VideoRecorder()
{
	__vid.ok = false;
	av_register_all();
}

VideoRecorder::~VideoRecorder()
{
	closeFile();
}

void VideoRecorder::openFile(const char *filename, int width, int height, int rate)
{
	if (__vid.ok)
		closeFile();
	
	
	// Detect format
	AVOutputFormat *fmt = guess_format(NULL, filename, NULL);
	if (!fmt)
	{
		std::cerr << "VideoRecorder::openFile : could not guess format, defaulting to MPEG" << std::endl;
		fmt = guess_format("mpeg", NULL, NULL);
	}
	if (!fmt)
	{
		std::cerr << "VideoRecorder::openFile : could not find suitable output format, aborting" << std::endl;
		return;
	}
	
	// Init format context
	__vid.oc.oformat = fmt;
	snprintf(__vid.oc.filename, sizeof(__vid.oc.filename), "%s", filename);
	
	// Create video stream
	__vid.vs = av_new_stream(&__vid.oc, 0);
	if (!__vid.vs)
	{
		std::cerr << "VideoRecorder::openFile : could not allocate video stream" << std::endl;
		return;
	}
	AVCodecContext *c = &__vid.vs->codec;
	c->codec_id = fmt->video_codec;
	c->codec_type = CODEC_TYPE_VIDEO;
	c->bit_rate = rate;
	c->width = width;
	c->height = height;
	c->frame_rate = 25;
	c->frame_rate_base = 1;
	c->gop_size = 12;
	if (c->codec_id == CODEC_ID_MPEG2VIDEO)
	{
		// just for testing, we also add B frames
		c->max_b_frames = 2;
	}
	if (c->codec_id == CODEC_ID_MPEG1VIDEO)
	{
		/* needed to avoid using macroblocks in which some coeffs overflow 
			this doesnt happen with normal video, it just happens here as the 
			motion of the chroma plane doesnt match the luma plane */
		c->mb_decision = 2;
	}
	// some formats want stream headers to be seperate
	if(!strcmp(__vid.oc.oformat->name, "mp4") || !strcmp(__vid.oc.oformat->name, "mov") || !strcmp(__vid.oc.oformat->name, "3gp"))
		c->flags |= CODEC_FLAG_GLOBAL_HEADER;
	
	av_set_parameters(&__vid.oc, NULL);
	
	// Find encoder
	AVCodec *codec = avcodec_find_encoder(c->codec_id);
	if (!codec)
	{
		std::cerr << "VideoRecorder::openFile : codec not found" << std::endl;
		return;
	}
	
	// Open the codec
	if (avcodec_open(c, codec) < 0)
	{
		std::cerr << "VideoRecorder::openFile : could not open codec" << std::endl;
		return;
	}
	
	// Allocate the memory for the frame to be encoded
	__vid.outbufSize = 200000;
	__vid.outbuf = (unsigned char *)malloc(__vid.outbufSize);
	__vid.toEncodeFrame = allocFrame(c->pix_fmt, c->width, c->height);
	if (!__vid.toEncodeFrame)
	{
		std::cerr << "VideoRecorder::openFile : could not allocate frame to be encoded" << std::endl;
		return;
	}
	
	// Open file
	if (!(fmt->flags & AVFMT_NOFILE))
		if (url_fopen(&__vid.oc.pb, filename, URL_WRONLY) < 0)
		{
			std::cerr << "VideoRecorder::openFile : could not open " << filename << ", aborting" << std::endl;
		}
	
	// Write header
	av_write_header(&__vid.oc);
	
	__vid.ok = true;
}

void VideoRecorder::encodeFrame(unsigned char *data)
{
	if (__vid.ok)
	{
		AVCodecContext *c = &__vid.vs->codec;
		
		AVPicture inputPict;
		avpicture_fill(&inputPict, data, PIX_FMT_RGBA32, c->width, c->height);
		
		if (img_convert((AVPicture *)__vid.toEncodeFrame, c->pix_fmt, &inputPict, PIX_FMT_RGBA32, c->width, c->height) < 0)
		{
			std::cerr << "Pixel format conversion failed" << std::endl;
			return;
		}
		
		//__vid.outbufSize = avcodec_encode_video(c, __vid.outbuf, __vid.outbufSize, __vid.toEncodeFrame);
		int outSize = avcodec_encode_video(c, __vid.outbuf, __vid.outbufSize, __vid.toEncodeFrame);

		int ret;
		if(outSize != 0)
		{
			AVPacket pkt;
			av_init_packet(&pkt);
			
			pkt.pts= c->coded_frame->pts;
			if(c->coded_frame->key_frame)
				pkt.flags |= PKT_FLAG_KEY;
			pkt.stream_index= __vid.vs->index;
			pkt.data = __vid.outbuf;
			pkt.size = outSize;
			
			// write the compressed frame in the media file
			ret = av_write_frame(&__vid.oc, &pkt);
		}
		else
		{
			ret = 0;
		}
		if (ret != 0)
			std::cerr << "Error while writing video frame" << std::endl;
	}
}



void VideoRecorder::closeFile()
{
	if (__vid.ok)
	{
		// TODO : free remaining data in codec
		// Close codec
		avcodec_close(&__vid.vs->codec);
		
		// Free temp picture memory
		av_free(__vid.toEncodeFrame->data[0]);
		av_free(__vid.toEncodeFrame);
		av_free(__vid.outbuf);
		
		// Write EOF
		av_write_trailer(&__vid.oc);
		
		// Free streams
		for(int i = 0; i < __vid.oc.nb_streams; i++)
			av_freep(&__vid.oc.streams[i]);
			
		// Free files
		if (!(__vid.oc.oformat->flags & AVFMT_NOFILE))
			url_fclose(&__vid.oc.pb);
		
		__vid.ok = false;
	}
}

#endif // HAVE_AVCODEC
