/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef EVOLUTION_H
#define EVOLUTION_H

#include <valarray>
#include <ishtar/network.h>

/*!	\file Evolution.h
	\brief Interfaces of abstract evolutions
*/

/*!	\defgroup geneticalgos Genetic Algorithms
	The genetic algorithms for evolutionary experiments.
*/

namespace Teem
{
	class GenotypeDecoder;
	class Experiment;
	
	//! An interface for serializable data
	//! \ingroup geneticalgos
	struct Serializable
	{
		//! Virtual destructor, doing nothing
		virtual ~Serializable() { }
		//! Load experiment datas from stream
		virtual void load(Ishtar::InputStream *stream) = 0;
		//! Save experiment datas to stream
		virtual void save(Ishtar::OutputStream *stream) = 0;
	};
	
	//! Array of genotype decoder
	//! \ingroup geneticalgos
	typedef std::valarray<GenotypeDecoder *> GenotypeDecoders;
	
	//! Interface to a class that implements a genotype decoder setup. Experiment classes typically derives from Setupable.
	//! \ingroup geneticalgos
    class Setupable
	{
	public:
		//! Virtual destructor, doing nothing
		virtual ~Setupable() { }
		//! Setup the decoders, in particular the IO related settings (input num, etc.)
		virtual void setupDecoders(GenotypeDecoders &decoders) = 0;
	};
	
	//! Interface for all GAs, both generational and steady-state
	class EvolutionRun
	{
	protected:
		//! Genome decoders used by the evolution run.
		GenotypeDecoders genotypeDecoders;
		
	protected:
		//! Delete genotype decoders
		void deleteGenotypeDecoders();
		
	public:
		//! Destructor
		virtual ~EvolutionRun();
		
		//! Create the necessary genotype decoders.
		virtual void createGenotypeDecoders(Setupable *experiment) = 0;
		
		//! Create a population of random genomes, using the genotype decoder
		virtual void createRandomPopulation() = 0;
		
		//! Load a most recent population. 
		virtual bool loadPopulation(Experiment *experiment) = 0 ;
		//! Load a population from a specific generation or time
		virtual bool loadPopulation(Experiment *experiment, const char *) = 0;
		
		//! Do the evolution
		virtual void run(Setupable *experiment) = 0;
		
		//! Test the best Individual and test it (used to view and evaluate indivuals for analysis)
		virtual void testBestIndividual(Setupable *experiment) = 0;
		//! Test a certain individual and evaluate it (used to view and evaluate individuals for analysis)
		virtual void testIndividual(Setupable *experiment, const char *) = 0;
	};
}

#endif
