/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __VALUE_GENOME_H
#define __VALUE_GENOME_H

#include "Genome.h"
#include <valarray>

/*!	\file ValueGenome.h
	\brief Interfaces of genomes based on array of values
*/

namespace Teem
{
	//! A genome which stores values
	//! \ingroup geneticalgos
	class ValueGenome: public Genome
	{
	public:
		//! Read a double value in range [min..max[ from the genome
		virtual double getDouble(double min, double max) = 0;
		//! Read a float value in range [min..max[ from the genome
		virtual float getFloat(float min, float max) = 0;
		//! Read a signed int value in range [min..max] from the genome using precision bits
		virtual int getSignedInt(unsigned precision, int min, int max) = 0;
		//! Read an unsigned int value in range [min..max] from the genome using precision bits
		virtual unsigned getUnsignedInt(unsigned precision, unsigned min, unsigned max) = 0;
		
		//! Set reading position to 0
		virtual void resetReading(void) = 0;
	};
	
	//! A genome which uses bits to store values
	//! \ingroup geneticalgos
	class BitGenome: public ValueGenome
	{
	protected:
		//! real datas of the genome
		std::valarray<unsigned> bits;
		//! length in bits of the genome
		size_t length;
		
		//! cached ishtar parameters
		Ishtar::Variable<double> mutationProbability;
		//! cached ishtar parameters
		Ishtar::Variable<unsigned> defaultPrecision;
		
		//! positiong for reading, internal machinery
		size_t readPos;
		
	protected:
		//! return true if word pos is within bound, returns false or assert otherwise depending on Genome::disableAssertion
		bool checkBoundWord(size_t pos);
		
	public:
		//! Constructor, deserialize from stream
		BitGenome(Ishtar::InputStream *stream);
		//! Constructor, size bits
		BitGenome(size_t size);
		
		//! Returns bit at pos
		bool getBit(size_t pos);
		//! [] operator to get bit at pos, convenient
		bool operator[](size_t pos) { return getBit(pos); }
		//! Set bit at pos
		void setBit(size_t pos, bool bit);
		//! Seel to pos
		void seek(size_t pos);
		
		//! Returns the length in bits
		size_t getLength() { return length; }
		
		//! Returns the number of unsigned in the bits storage. Usefull for specific implementations.
		size_t getWordCount() { return bits.size(); }
		//! Directly set an unsigned in the bits storage. Usefull for specific implementations.
		void setWord(size_t pos, unsigned val) { if (checkBoundWord(pos)) bits[pos] = val; }
		//! Directly get an unsigned in the bits storage. Usefull for specific implementations.
		unsigned getWord(size_t pos) { if (checkBoundWord(pos)) return bits[pos]; }
		
		//! Get a value of type unsigned int on range [0..2^precision-1] by directly reading bit per bit, increment reading position
		unsigned getUnsignedIntRaw(unsigned precision)
		{
			
			unsigned int val = 0;
			for (unsigned i=0; i<precision; i++)
				val = getBit(readPos++) ? (val + (1<<i)) : val;
			return val;
		}
		
		//! Get a value of type T on range [0..1] using bitCount bits, increment reading position
		template<typename T>
		T getValue(unsigned bitCount)
		{
			if (bitCount <= 32)
			{
				unsigned int val = 0;
				for (unsigned i=0; i<bitCount; i++)
					val = getBit(readPos++) ? (val + (1<<i)) : val;
				return (static_cast<T>(val) / static_cast<T>((1<<bitCount)-1));
			}
			else
			{
				std::cerr << "BitGenome::getValue<T> : deserializing value bigger than 32 bits not supported" << std::endl;
				return 0;
			}
		}
		
		//! Get a value of type T on range [min..max] using bitCount bits, increment reading position
		template<typename T>
		T getValueRange(unsigned bitCount, T min, T max)
		{
			return min + (getValue<T>(bitCount)*(max - min));
		}
		
		//! Get a value of type double on range [min..max] using defaultPrecision bits, increment reading position
		virtual double getDouble(double min, double max)
		{
			return getValueRange<double>(defaultPrecision, min, max);
		}
		
		//! Get a value of type float on range [min..max] using defaultPrecision bits, increment reading position
		virtual float getFloat(float min, float max)
		{
			return getValueRange<float>(defaultPrecision, min, max);
		}
		
		//! Get a value of type signed int on range [min..max] using precision bits, increment reading position
		virtual int getSignedInt(unsigned precision, int min, int max)
		{
			return static_cast<unsigned>(getValueRange<double>(precision, static_cast<double>(min), static_cast<double>(max)));
		}
		
		//! Get a value of type unsigned int on range [min..max] using precision bits, increment reading position
		virtual unsigned getUnsignedInt(unsigned precision, unsigned min, unsigned max)
		{
			return static_cast<unsigned>(getValueRange<double>(precision, static_cast<double>(min), static_cast<double>(max)));
		}
		
		virtual void resetReading(void);
		
		//! Number of bits used for floats and doubles
		void setDefaultPrecision(unsigned p) { defaultPrecision = p; }
		
		virtual void randomize(void);
		virtual void mutate(void);
		virtual void cross(const Genome *other);
		virtual Genome *clone(void);
		
		virtual void save(Ishtar::OutputStream *stream);
	};
}

#endif
