/*
  Copyright (C) 2001-2005 Stephane Magnenat & Luc-Olivier de Charrière
  for any question or comment contact us at nct@ysagoon.com or nuage@ysagoon.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "SDLFont.h"
#include <assert.h>
#include <SDL_image.h>

/*!	\file SDLFont.cpp
	\brief Implementation of fonts using SDL
*/

namespace GAG
{
	//! Source-Destination-Clipping
	void sdcRects(SDL_Rect *source, SDL_Rect *destination, const SDL_Rect &clipping)
	{
		//sdc= Source-Destination-Clipping
		//Use if destination have the same size than source & cliping on destination
		int dx=clipping.x-destination->x;
		int dy=clipping.y-destination->y;
	
		int sw=source->w;
		int sh=source->h;
	
		if (dx>0)
		{
			source->x+=dx;
			destination->x+=dx;
	
			sw-=dx;
			destination->w-=dx;
		}
		if (dy>0)
		{
			source->y+=dy;
			destination->y+=dy;
	
			sh-=dy;
			destination->h-=dy;
		}
	
		int dwx=(destination->x+destination->w)-(clipping.x+clipping.w);
		int dhy=(destination->y+destination->h)-(clipping.y+clipping.h);
	
		if (dwx>0)
		{
			sw-=dwx;
			destination->w-=dwx;
		}
		if (dhy>0)
		{
			sh-=dhy;
			destination->h-=dhy;
		}
	
		if (sw>0)
			source->w=(Uint16)sw;
		else
			source->w=0;
	
		if (sh>0)
			source->h=(Uint16)sh;
		else
			source->h=0;
	}
	
	
	SDLFont::SDLFont()
	{
		font = NULL;
		
	}
	
	SDLFont::~SDLFont()
	{
		if (font)
			TTF_CloseFont(font);
			
	}
	
	bool SDLFont::load(const char *filename, unsigned size)
	{
		SDL_RWops *fontStream = SDL_RWFromFile(filename, "rb");
		if (fontStream)
		{
			font = TTF_OpenFontRW(fontStream, 1, size);
			if (font)
			{
				setColor(0, 0, 0, DrawableSurface::ALPHA_OPAQUE);
				setStyle(Font::STYLE_NORMAL);
				return true;
			}
		}
		return false;
	}
	
	int SDLFont::getStringWidth(const char *string) const
	{
		int w, h;
		TTF_SizeUTF8(font, string, &w, &h);
		return w;
	}
	
	int SDLFont::getStringHeight(const char *string) const
	{
		if (string)
		{
			int w, h;
			TTF_SizeUTF8(font, string, &w, &h);
			return h;
		}
		else
		{
			return TTF_FontHeight(font);
		}
	}
	
	bool SDLFont::printable(char c) const
	{
		return (c>=32);
	}
	
	void SDLFont::setColor(Uint8 r, Uint8 g, Uint8 b, Uint8 a)
	{
		assert(font);
		
		while (colorStack.size() > 0)
			colorStack.pop();
		
		pushColor(r, g, b, a);
	}
	
	void SDLFont::pushColor(Uint8 r, Uint8 g, Uint8 b, Uint8 a)
	{
		assert(font);
		
		SDL_Color c;
		c.r = r;
		c.g = g;
		c.b = b;
		c.unused = a;
		colorStack.push(c);
	}
	
	void SDLFont::popColor(void)
	{
		assert(font);
		
		if (colorStack.size() > 1)
			colorStack.pop();
	}
	
	void SDLFont::getColor(Uint8 *r, Uint8 *g, Uint8 *b, Uint8 *a) const
	{
		assert(font);
		
		SDL_Color c;
		c = colorStack.top();
		*r = c.r;
		*g = c.g;
		*b = c.b;
		*a = c.unused;
	}
	
	void SDLFont::setStyle(unsigned style)
	{
		assert(font);
		
		while (styleStack.size() > 0)
			styleStack.pop();
			
		pushStyle(style);
	}
	
	void SDLFont::pushStyle(unsigned style)
	{
		assert(font);
		
		styleStack.push(style);
		TTF_SetFontStyle(font, style);
	}
	
	void SDLFont::popStyle(void)
	{
		assert(font);
		
		if (styleStack.size() > 1)
		{
			styleStack.pop();
			TTF_SetFontStyle(font, styleStack.top());
		}
	}
	
	unsigned SDLFont::getStyle(void) const
	{
		assert(font);
		
		return styleStack.top();
	}
	
	void SDLFont::drawString(SDL_Surface *Surface, int x, int y, int w, const char *text, SDL_Rect *clip) const
	{
		assert(text);
		assert(font);
		assert(colorStack.size()>0);
		
		SDL_Color c = colorStack.top();
		SDL_Surface *s;
		s=TTF_RenderUTF8_Blended(font, text, c);
		if (s == NULL)
			return;
	
		SDL_Rect sr;
		sr.x=0;
		sr.y=0;
		sr.w=s->w;
		sr.h=s->h;
	
		SDL_Rect r;
		r.x=x;
		r.y=y;
		if (w)
			r.w=w;
		else
			r.w=s->w;
		r.h=s->h;
	
		SDL_Rect oc;
		if (clip)
		{
			SDL_GetClipRect(Surface, &oc);
			sdcRects(&sr, &r, *clip);
			SDL_SetClipRect(Surface, &r);
		}
	
		SDL_BlitSurface(s, &sr, Surface, &r);
	
		if (clip)
		{
			SDL_SetClipRect(Surface, &oc);
		}
		
		SDL_FreeSurface(s);
	}
}
