#ifndef __LOGGER
#define __LOGGER

#include <ishtar/settings.h>
#include <ostream>
#include <fstream>
#include <vector>

/*!	\file Logger.h
	\brief Interfaces of logging helper classes
*/

/*!	\defgroup utils Utilities
	Helper classes usefull in other parts of Teem
*/

namespace Teem
{
	//! Abstract class for a generic logger.
	//! \ingroup utils
	class Logger
	{
	public:
	
		//! Constructor.
		Logger(const char* fileName) : logFile(fileName), inited(false) { }
		//! Destructor
		virtual ~Logger();
		
		//! Add a reference to the logger.
		template<typename T> void addVariable(const std::string &label, const T &ref) { add(label, new LoggableVariable<T>(ref)); }
		
		//! Add an Ishtar variable to the logger.
		template<typename T> void addIshtarVariable(const std::string &label, const char* id) { add(label, new LoggableIshtarVariable<T>(id)); }
		
		//! Add a collection of variables using iterators
		template<typename ForwardIterator> void addVariables(const std::string &label, ForwardIterator from, ForwardIterator to) { while(from != to) { addVariable(label, *from); ++from; } }
		
		//! Add a new snapshot to the log file.
		void step();
	
	protected:
	
		//! Abstract class for loggable unit.
		//! This is a conceptual point to some numeric entity that can be dumped to some output stream.
		//! Subclass must support copy operations.
		class Loggable
		{
		public:
			//! Virtual destructor, do nothing
			virtual ~Loggable() { }
			//! Log unit to output stream s
			virtual void print(std::ostream &s) const = 0;
		};
		
		//! Loggable based on pointer. This class save a pointer to some variable
		//! and dump it by dereferencing the pointer.
		template<typename T> class LoggableVariable : public Loggable
		{
		public:
			//! Constructor, log variable r
			LoggableVariable(const T &r) : ref(r) { }
			virtual void print(std::ostream &s) const { s << ref; }
		protected:
			const T& ref; //!< reference to variable to be logged
		};
		
		//! Lagable based on an ishtar variable.
		template<typename T> class LoggableIshtarVariable :  public Loggable
		{
		public:
			//! Constructor, log an ishtar variable named id
			LoggableIshtarVariable(const char* id) : variable(id) { }
			virtual void print(std::ostream &s) const { s << variable; }
		protected:
			Ishtar::Variable<T> variable; //!< variable to be logged
		};
		
		friend std::ostream& operator<< (std::ostream &os, const Logger::Loggable &loggable);
	
		//! Map for all logable units
		typedef std::map<std::string, std::vector<Loggable*> > LoggableMap;
		std::ofstream logFile; //!< file to log to
		bool inited; //!< true if log is initialised. Used for delayed initialisation
		LoggableMap loggables; //!< all logables units
		
		//! Add a new loggable in the log under a certain label.
		void add(const std::string &label, Loggable* loggable);
		
		//! Init the log file (e.g. write information in the begining of the log file).
		virtual void initLog(std::ostream &log) = 0;
		//! Dump a snapshot (e.g. write the content of all variables to the log file).
		virtual void dumpSnapshot(std::ostream &log) = 0;
	};
	
	//! Logger that use a format that is Octave-friendly.
	//! \ingroup utils
	class OctaveLogger : public Logger
	{
	public:
		//! Constructor.
		//! @param fileName name to use for the log file
		//! @param matlab use matlab mode instead of octave mode (load_log vs. teem_log)
		OctaveLogger(const char* fileName, bool matlab = false) : Logger(fileName), matlabMode(matlab) { }
		//! Destructor.
		virtual ~OctaveLogger() { }
	
	protected:
	
		const bool matlabMode; //!< Be compatible with matlab's load_log (instead of octave's teem_log)
	
		//! Init the log file (e.g. write information in the begining of the log file).
		virtual void initLog(std::ostream &log);
		//! Dump a snapshot (e.g. write the content of all variables to the log file).
		virtual void dumpSnapshot(std::ostream &log);
	};
	
	//! Operator << for ostreams and Loggables.
	//! \ingroup utils
	inline std::ostream& operator<< (std::ostream &os, const Logger::Loggable &loggable)
	{
		loggable.print(os);
		return os;
	}
}

#endif
