#ifndef __REGISTRAR_H
#define __REGISTRAR_H

#include <iostream>
#include <map>
#include <string>
#include <assert.h>

/*!	\file Registrar.h
	\brief Templatised registry helper classes
*/

namespace Teem
{
	/**
		A Registrar for classes with two parameters constructors
		
		\li T is the base class
		\li IdT is the type of ID
		\li Sub is a particular subclass to be registered
		
		\b Usage
		
		A, is a class; B, C are its subclasses
\code
Registrar<A, std::string, int, int> reg;

reg.registerItem<B>("you");
reg.registerItem<C>("yeah");

A* a1 = reg.createItem("you", 0, 0);
A* a2 = reg.createItem("yeah", 0, 0);
\endcode
		
		\ingroup utils
	*/
	template<typename T, typename IdT, typename Arg1T = void, typename Arg2T = void> class Registrar
	{
	public:
		//! Register a class by storing a pointer to its constructor
		template<typename Sub> void registerItem(const IdT &id)
		{
			mItems[id] = &Creator<Sub>;
		}
		
		//! Create an instance of a registered class and pass arg1 and arg2 to its constructor
		T* createItem(const IdT &id, Arg1T arg1, Arg2T arg2)
		{
			if (mItems.find(id) == mItems.end())
			{
				std::cerr << "Class of type " << id << " unregistred, aborting.\n\nKnown classes are:\n";
				dump(std::cerr);
				std::cerr << std::endl;
				exit(1);
			}
			return mItems[id](arg1, arg2);
		}
		
		//! Print the list of registered classes to stream
		void dump(std::ostream &stream)
		{
			for (typename std::map<IdT, CreatorFunc>::iterator it = mItems.begin(); it != mItems.end(); ++it)
				stream << "* " << it->first << "\n";
		}
	
	protected:
		//! Pointer to constructor of class T
		typedef T* (*CreatorFunc)(Arg1T, Arg2T);
		std::map<IdT, CreatorFunc> mItems; //!< registered constructors
		
		//! Create a new class of type Sub with arguments arg1 and arg2
		template<typename Sub> static T* Creator(Arg1T arg1, Arg2T arg2)
		{
			return new Sub(arg1, arg2);
		}
	};
	
	/**
		A Registrar for classes with one parameter constructors
		
		\li T is the base class
		\li IdT is the type of ID
		\li Sub is a particular subclass to be registered
		
		\b Usage
		
		A, is a class; B, C are its subclasses
\code
Registrar<A, std::string, int> reg;

reg.registerItem<B>("you");
reg.registerItem<C>("yeah");

A* a1 = reg.createItem("you", 0);
A* a2 = reg.createItem("yeah", 0);
\endcode
		
		\ingroup utils
	*/
	template<typename T, typename IdT, typename ArgT> class Registrar<T, IdT, ArgT, void>
	{
	public:
		//! Register a class by storing a pointer to its constructor
		template<typename Sub> void registerItem(const IdT &id)
		{
			mItems[id] = &Creator<Sub>;
		}
		
		//! Create an instance of a registered class and pass arg to its constructor
		T* createItem(const IdT &id, ArgT arg)
		{
			if (mItems.find(id) == mItems.end())
			{
				std::cerr << "Class of type " << id << " unregistred, aborting.\n\nKnown classes are:\n";
				dump(std::cerr);
				std::cerr << std::endl;
				exit(1);
			}
			return mItems[id](arg);
		}
		
		//! Print the list of registered classes to stream
		void dump(std::ostream &stream)
		{
			for (typename std::map<IdT, CreatorFunc>::iterator it = mItems.begin(); it != mItems.end(); ++it)
				stream << "* " << it->first << "\n";
		}
		
	protected:
		//! Pointer to constructor of class T
		typedef T* (*CreatorFunc)(ArgT);
		std::map<IdT, CreatorFunc> mItems; //!< registered constructors
		
		//! Create a new class of type Sub with argument arg
		template<typename Sub> static T* Creator(ArgT arg)
		{
			return new Sub(arg);
		}
	};
	
	/**
		A Registrar for classes with no parameter constructors
		
		\li T is the base class
		\li IdT is the type of ID
		\li Sub is a particular subclass to be registered
		
		\b Usage
		
		A, is a class; B, C are its subclasses
\code
Registrar<A, std::string> reg;

reg.registerItem<B>("you");
reg.registerItem<C>("yeah");

A* a1 = reg.createItem("you");
A* a2 = reg.createItem("yeah");
\endcode
		
		\ingroup utils
	*/
	template<typename T, typename IdT> class Registrar<T, IdT, void, void>
	{
	public:
		//! Register a class by storing a pointer to its constructor
		template<typename Sub> void registerItem(const IdT &id)
		{
			mItems[id] = &Creator<Sub>;
		}
		
		//! Create an instance of a registered class
		T* createItem(const IdT &id)
		{
			if (mItems.find(id) == mItems.end())
			{
				std::cerr << "Class of type " << id << " unregistred, aborting.\n\nKnown classes are:\n";
				dump(std::cerr);
				std::cerr << std::endl;
				exit(1);
			}
			return mItems[id]();
		}
		
		//! Print the list of registered classes to stream
		void dump(std::ostream &stream)
		{
			for (typename std::map<IdT, CreatorFunc>::iterator it = mItems.begin(); it != mItems.end(); ++it)
				stream << "* " << it->first << "\n";
		}
		
	protected:
		//! Pointer to constructor of class T
		typedef T* (*CreatorFunc)();
		std::map<IdT, CreatorFunc> mItems; //!< registered constructors
		
		//! Create a new class of type Sub
		template<typename Sub> static T* Creator()
		{
			return new Sub();
		}
	};
}

#endif
