/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __TEEM_ENKI_H
#define __TEEM_ENKI_H

#include <ishtar/settings.h>
#include <an/Geometry.h>
#include <an/Random.h>

#include "World.h"
#include "Entity.h"
#include "../enkiviewer/EnkiViewer.h"
#include "../utils/Registrar.h"

/*!	\file Enki.h
	\brief World interface to Enki
*/

namespace Enki
{
	class World;
	class Alice;
	class PhysicalObject;
	class FeedableSbot;
	class SbotActiveObject;
	class Khepera;
}

namespace Teem
{
	class EnkiSbot;
	class EnkiSbotActiveObject;
	class EnkiKhepera;
	class EnkiAlice;
	class EnkiAliceSmallDisk;
	class EnkiAliceLargeDisk;
	class EnkiRectangularBlock;
	class EnkiRealisticConditions;
	#ifndef HAVE_VIEWER
	class EnkiViewer;
	#endif

	//! A container for noise conditions.
	//! \ingroup world
	class EnkiRealisticConditions
	{
	public:
		double motorInstantNoise(); //!< Changes every time step. To be added to each wheel.
		double motorCoeff(); //!<  Changes every epoch. Proportional to the speed, to be added to each wheel.
		double camNoise(); //!< Gaussian noise to be added to each ray of the cam.
		bool   camBadPixel(); //!< Return true every x calls on average, according to camProbBadPixel

	protected:
		friend class EnkiWorld;
		//! Create the random generators.
		EnkiRealisticConditions();
		//! Destructor.
		~EnkiRealisticConditions();

		Ishtar::Variable<double> motorInstantNoiseMean; //!< mean of instant gaussian motor noise
		Ishtar::Variable<double> motorInstantNoiseStd; //!< standard deviation of instant gaussian motor noise
		Ishtar::Variable<double> motorCoeffRange; //!< range for motor coefficient (constant error). Coefficient will be uniformly distributed in [1+motorCoeffRange..1-motorCoeffGenerator]
		Ishtar::Variable<double> camNoiseMean; //!< mean of gaussian camera noise
		Ishtar::Variable<double> camNoiseStd; //!< standard deviation of gaussian camera noise
		Ishtar::Variable<double> camProbBadPixel; //!< probability of bad pixel in camera

		An::UniformRand motorCoeffGenerator; //!< generator for motor coefficient (constant error)
	};


	//! This is the interface to Enki, our fast 2D world simulator.
	//! \ingroup world
	class EnkiWorld : public SquareArenaWorld, public SbotCreator, public SbotActiveObjectCreator, public KheperaCreator, public AliceCreator, public AliceSmallDiskCreator, public AliceLargeDiskCreator, public RectangularBlockCreator
	{
	protected:
		friend class EnkiKhepera;
		friend class EnkiAlice;
		friend class EnkiAliceSmallDisk;
		friend class EnkiAliceLargeDisk;
		friend class EnkiSbot;
		friend class EnkiSbotActiveObject;
		friend class EnkiRectangularBlock;

		Enki::World *world; //!< Instance of Enki world
		EnkiViewer *viewer; //!< Instance of Enki viewer
	    Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the EnkiWorld.
		Ishtar::Variable<unsigned, true, 2> dimension; //!< Dimensions of Enki world
		Ishtar::Variable<unsigned> useVideo; //!< True if we record videos.
		Ishtar::ArrayHelper objectsArrayHelper; //!< A helper to get the root for PhysicalObject in objects.
		Ishtar::Variable<bool> realisticMode; //!< If true, postprocessing is applied to inputs/outputs so that simulation matches reality more closely.
		std::vector<Entity *> autoCreatedEntities; //!< Entities that have been automatically created in constructor from an Ishtar. settings file
		EnkiRealisticConditions realisticConditions; //!< Instance of EnkiRealisticConditions

	public:
		//! Constructor
		EnkiWorld(bool createEnkiViewer = true);
		//! Destructor: Delete viewer, video, automatically created entities and world.
		virtual ~EnkiWorld();

		// hack to create viewer
		#ifdef HAVE_VIEWER
		//! Create the viewer.
		template<class ViewerT>
		void createViewer() { viewer = new ViewerT(world, useVideo != 0); viewer->startRecordingVideo("enki.avi"); }
		//! Create the viewer.
		template<class ViewerT, class T1>
		void createViewer(T1 p1) { viewer = new ViewerT(world, useVideo != 0, p1); viewer->startRecordingVideo("enki.avi"); }
		#endif

		//! Create an sbot.
		virtual Sbot *createSbot();
		//! Create an active object (sbot).
		virtual SbotActiveObject *createSbotActiveObject();
		//! Create a Khepera
		virtual Khepera *createKhepera();
		//! Create an Alice
		virtual Alice *createAlice();
		//! Create an Alice small disk.
		virtual AliceLargeDisk *createAliceLargeDisk();
		//! Create an Alice large disk.
		virtual AliceSmallDisk *createAliceSmallDisk();
		//! Create a rectangular block.
		virtual RectangularBlock *createRectangularBlock();

		// implementation of World
		virtual void step(double dt);

		// implementation of SquareArenaWorld
		virtual An::Point getDimension();
		virtual void setWallTextures(const std::valarray<An::Texture> &textures);

		// ! Use new noises conditions.
		virtual void useWalls(bool use);

		//! Return the instance of the world in Enki.
		Enki::World *getEnkiInstance() { return world; }
	};

	//! A positionable object in Enki.
	//! \ingroup world
	template<typename T>
	class EnkiPositionable : virtual public Positionable
	{
	public:
		T *object; //!< Pointer to object instance in Enki.

	public:
		// implementation of Positionable
		virtual void setOrientation(double angle) {  object->angle = angle; }
		virtual void setPosition(const An::Point &pos) { object->pos = pos; }
		virtual const An::Point &getPosition(void) { return object->pos; }
		virtual double getOrientation(void) { return object->angle; }

		//! Return the instance of the object in Enki.
		T *getEnkiInstance(void) { return object; }
	};

	//! This is the implementation of a Khepera in the Enki simulator.
	//! \ingroup world
	class EnkiKhepera : public EnkiPositionable<Enki::Khepera>, public Khepera
	{
	private:
		friend class EnkiWorld;
		friend class Registrar<Entity, std::string, EnkiWorld *, std::string>;
		EnkiWorld *world; //!< Pointer to world to be able to remove object in destructor.
		Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the Enki object.

		//! Private constructor, created from world.
		EnkiKhepera(EnkiWorld *world, const std::string &root);

	public:
		//! Destructor, remove object from world and destroy Enki instance.
		virtual ~EnkiKhepera();

		// implementation of DifferentialWheeled
		virtual void setSpeed(double left, double right);
		virtual void getEncoders(double &left, double &right);

		// implementation of IREquipped
		virtual double getIRSensor(unsigned index);

		// implementation of CircularCameraEquipped
		virtual An::Color getCameraPixel(unsigned index);
		virtual unsigned getCameraSize();

		// reimplementation of EnkiPositionable to reset encoders
		virtual void setPosition(const An::Point &pos);
	};

	//! This is the implementation of an Alice in the Enki simulator.
	//! \ingroup world
	class EnkiAlice : public EnkiPositionable<Enki::Alice>, public Alice
	{
	private:
		friend class EnkiWorld;
		friend class Registrar<Entity, std::string, EnkiWorld *, std::string>;
		EnkiWorld *world; //!< Pointer to world to be able to remove object in destructor.
		Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the Enki object.
		
		//! Private constructor, created from world.
		EnkiAlice(EnkiWorld *world, const std::string &root);

	public:
		//! Destructor, remove object from world and destroy Enki instance.
		virtual ~EnkiAlice();

		// implementation of DifferentialWheeled
		virtual void setSpeed(double left, double right);

		// implementation of IRSensorEquipped
		virtual double getIRSensor(unsigned index);

		// implementation of Alice
		virtual void setTransmitValue(unsigned val);
		virtual void setTransmit(bool enabled);
		virtual bool wasCommunication(void);
		virtual unsigned getSensorId(void);
		virtual unsigned getSensorValue(void);
		virtual void getSpeed(double &left, double &right);

		// implementation of Camera1DEquipped
		virtual An::Color getCameraPixel(unsigned index);
		virtual unsigned getCameraSize();
	};

	//! This is the implementation of a small Alice disk in the Enki simulator.
	//! \ingroup world
	class EnkiAliceSmallDisk : public EnkiPositionable<Enki::PhysicalObject>, public AliceSmallDisk
	{
	private:
		friend class EnkiWorld;
		friend class Registrar<Entity, std::string, EnkiWorld *, std::string>;
		EnkiWorld *world; //!< Pointer to world to be able to remove object in destructor.
		Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the Enki object.

		//! Private constructor, created from world.
		EnkiAliceSmallDisk(EnkiWorld *world, const std::string &root);

	public:
		//! Destructor, remove object from world and destroy Enki instance.
		virtual ~EnkiAliceSmallDisk();
	};

	//! This is the implementation of a small Alice disk in the Enki simulator.
	//! \ingroup world
	class EnkiAliceLargeDisk : public EnkiPositionable<Enki::PhysicalObject>, public AliceLargeDisk
	{
	private:
		friend class EnkiWorld;
		friend class Registrar<Entity, std::string, EnkiWorld *, std::string>;
		EnkiWorld *world; //!< Pointer to world to be able to remove object in destructor.
		Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the Enki object.

		//! Private constructor, created from world.
		EnkiAliceLargeDisk(EnkiWorld *world, const std::string &root);

	public:
		//! Destructor, remove object from world and destroy Enki instance.
		virtual ~EnkiAliceLargeDisk();
	};


	//! This is the implementation of an Sbot in the Enki simulator.
	//! \ingroup world
	class EnkiSbot : public EnkiPositionable<Enki::FeedableSbot>, public Sbot
	{
	private:
		friend class EnkiWorld;
		friend class Registrar<Entity, std::string, EnkiWorld *, std::string>;
		EnkiWorld *world; //!< Pointer to world to be able to remove object in destructor.
		Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the Enki object.

		//! If true, postprocessing is applied to inputs/outputs so that simulation matches reality more closely.
		static Ishtar::Variable<bool>* realistic;

		//! Private constructor, created from world.
		EnkiSbot(EnkiWorld *world, const std::string &root);

	public:
		//! Destructor, remove object from world and destroy Enki instance.
		virtual ~EnkiSbot();

		// implementation of DifferentialWheeled
		virtual void setSpeed(double left, double right);

		// implementation of Sbot
		virtual void getCameraAverage(An::Color meanPixels[8]);
		virtual void getCamera(An::Color meanPixels[128]);
		virtual void setRingColor(const An::Color &color);
		virtual void clearEnergy(void);
		virtual void setEnergy(double energy);
		virtual double getEnergy(void);
		virtual double getDEnergy(void);
		virtual void setFrequencies(unsigned frequencies);
		virtual unsigned getFrequencies(void);
		//! Get new noisy coefficient for wheel motors in realistic mode
		virtual void randomizeRealisticConditions(void);

	protected:
		double motorCoeffL; //!< left wheel motor noisy coefficient in realistic mode
		double motorCoeffR; //!< right wheel motor noisy coefficient in realistic mode
	};

	//! This is the implementation of an Sbot active object in the Enki simulator.
	//! \ingroup world
	class EnkiSbotActiveObject : public EnkiPositionable<Enki::SbotActiveObject>, public SbotActiveObject
	{
	private:
		friend class EnkiWorld;
		friend class Registrar<Entity, std::string, EnkiWorld *, std::string>;
		EnkiWorld *world; //!< Pointer to world to be able to remove object in destructor.
		Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the Enki object.
		//! Private constructor, created from world.
		EnkiSbotActiveObject(EnkiWorld *world, const std::string &root);

	public:
		//! Destructor, remove object from world and destroy Enki instance.
		virtual ~EnkiSbotActiveObject();

		// implementation of SbotActiveObject
		virtual void setEnergyProperties(double activeDuration, double inactiveDuration, bool consumeEnergy);
		virtual void setEnergyTime(double time);
		virtual void setEnergyFlow(double activeFlow, double inactiveFlow);
		virtual void setEnergy(double energy);
		virtual double getEnergy(void);
		virtual void setColor(const An::Color &activeColor, const An::Color &inactiveColor);
		virtual double getInteractionRange(void);
	};

	//! This is the implementation of a piece of static matter in the Enki simulator.
	//! \ingroup world
	class EnkiRectangularBlock : public EnkiPositionable<Enki::PhysicalObject>, public RectangularBlock
	{
	private:
		friend class EnkiWorld;
		friend class Registrar<Entity, std::string, EnkiWorld *, std::string>;
		EnkiWorld *world; //!< Pointer to world to be able to remove object in destructor.
		An::Polygone boundingSurface; //!< Bounding surface of this block, may be modified through Ishtar.
		Ishtar::VariablesPublisher ishtarPublisher; //!< Ishtar publisher for accessible variables of the Enki object.

		//! Private constructor, created from world.
		EnkiRectangularBlock(EnkiWorld *world, const std::string &root);

	public:
		//! Destructor, remove object from world and destroy Enki instance.
		virtual ~EnkiRectangularBlock();

		// implementation of RectangularBlock
		virtual void set(double w, double h, double x, double y, double a);
	};
}

#endif
