/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __TEEM_ENTITY_INTERFACES_H
#define __TEEM_ENTITY_INTERFACES_H

/*!	\file EntityInterfaces.h
	\brief Generic interfaces for entities
*/

namespace Teem
{
	//! This is the interface to any physical object (disks, robots, vacuum cleaners...)
	//! \ingroup world
	class Entity
	{
	public:
		//! Virtual destructor, do nothing
		virtual ~Entity() { }
	};

	//! This is the interface to anything that behaves like a two wheeled robot.
	//! \ingroup world
	class DifferentialWheeled
	{
	public:
		//! Virtual destructor, do nothing
		virtual ~DifferentialWheeled() { }
		
		//! Set the speed of the left and right wheels of the robot
		virtual void setSpeed(double left, double right) = 0;
		//virtual void getEncoders(double &left, double &right) = 0;
	};
	
	//! This is the interface to anything that can be precisely placed.
	//! \ingroup world
	class Positionable
	{
	public:
		//! Virtual destructor, do nothing
		virtual ~Positionable() { }
		
		//! Set the orientation of the object
		virtual void setOrientation(double angle) = 0;
		//! Set the position of the object
		virtual void setPosition(const An::Point &pos) = 0;
		//! Return the position of the object
		virtual const An::Point &getPosition(void) = 0;
		//! Return the orientation of the object
		virtual double getOrientation(void) = 0;
		
		//! Set the position and the orientation of the object
		virtual void setPosition(const An::Point &pos, double angle) { setPosition(pos); setOrientation(angle); }
	};
	
	//! Interface to anything with IR sensors.
	//! \ingroup world
	class IRSensorEquipped
	{
	public:
		//! Virtual destructor, do nothing
		virtual ~IRSensorEquipped() { }
		
		//! Return the value of IR sensor index
		virtual double getIRSensor(unsigned index) = 0;
	};
	
	//! Interface to anything equipped with a 1D camera. 
	//! \ingroup world
	class Camera1DEquipped
	{
	public:
		//! Virtual destructor, do nothing
		virtual ~Camera1DEquipped() { }
		
		//! Return the number of pixel in the camera
		virtual unsigned getCameraSize() = 0;
		//! return the value of camera pixel index
		virtual An::Color getCameraPixel(unsigned index) = 0;
	};
}


#endif
